/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.data;

import java.util.Collection;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

import vispaca.event.DestroyEmitter;
import vispaca.event.DestroyListener;
import vispaca.event.UpdateEmitter;
import vispaca.event.UpdateListener;


/**
 * Provides east access to data tables and selections.
 */
public class DataManager implements UpdateEmitter, DestroyEmitter, DestroyListener {

	private DataTable dataTable;

	private Set<DataElement> touched = new HashSet<>();
	private Set<DataElement> highlights = new HashSet<>();
	private Set<DataSelection> filters = new HashSet<>();

	private UpdateListener selectionUpdateListener = new SelectionUpdateListener();

	private Set<UpdateListener> updateListeners = new HashSet<>();
	private Set<DestroyListener> destroyListeners = new HashSet<>();

	private boolean destroyed = false;


	public DataManager(DataTable table) {
		setDataTable(table);
	}


	/**
	 * @return The data table of the visualization.
	 */
	public DataTable getDataTable() {
		return dataTable;
	}


	/**
	 * Sets the data table to use for the visualization.
	 *
	 * @param table The data table to use.
	 */
	protected void setDataTable(DataTable table) {
		Objects.requireNonNull(table);
		this.dataTable = table;
	}


	/**
	 * Adds a new filter for globally filtering out elements.
	 *
	 * @param filter The filter to add.
	 * @return True if the filter was added, false if it was already present.
	 */
	public boolean addFilter(DataSelection filter) {
		Objects.requireNonNull(filter);
		boolean updated = filters.add(filter);

		if (updated) {
			filter.addUpdateListener(selectionUpdateListener);
			fireUpdateListeners();
		}

		return updated;
	}


	/**
	 * Removes a filter.
	 *
	 * @param filter The filter to remove.
	 * @return True if the filter was present.
	 */
	public boolean removeFilter(DataSelection filter) {
		Objects.requireNonNull(filter);
		boolean updated = filters.remove(filter);

		if (updated) {
			filter.removeUpdateListener(selectionUpdateListener);
			fireUpdateListeners();
		}

		return updated;
	}


	/**
	 * @return All globally present filters.
	 */
	public DataSelection[] getFilters() {
		return filters.toArray(new DataSelection[] {});
	}


	/**
	 * Decides (using filters) whether or not an element is visible.
	 *
	 * @param element The element to check.
	 * @return True if the element is visible, false if it was filtered out.
	 */
	public boolean isVisible(DataElement element) {
		for (DataSelection filter : filters) {
			if (filter.isSelected(element)) {
				return false;
			}
		}

		return true;
	}


	/**
	 * Decides whether or not an element should be highlighted.
	 *
	 * @param element The element to check.
	 * @return True if the element is highlighted.
	 */
	public boolean isHighlighted(DataElement element) {
		return highlights.contains(element);
	}


	/**
	 * Highlights elements.
	 *
	 * @param elements The elements to be highlighted.
	 * @return True if the highlights changed.
	 */
	public boolean addHighlight(DataElement element) {
		boolean updated = highlights.add(element);

		if (updated) {
			fireUpdateListeners();
		}

		return updated;
	}


	/**
	 * Causes an element not to be highlighted anymore.
	 *
	 * @param element The element which should not be highlighted.
	 * @return True if the element was highlighted
	 */
	public boolean removeHighlight(DataElement element) {
		boolean updated = highlights.remove(element);

		if (updated) {
			fireUpdateListeners();
		}

		return updated;
	}


	/**
	 * Causes all elements not to be highlighted anymore.
	 *
	 * @return True if the highlights were changed.
	 */
	public boolean clearHighlight() {
		if (highlights.isEmpty()) {
			return false;
		} else {
			highlights.clear();
			fireUpdateListeners();
			return true;
		}
	}


	/**
	 * Causes elements in the collection to switch highlight state.
	 *
	 * @param elements The elements to switch.
	 * @return True if the highlights were changed.
	 */
	public boolean swapHighlight(Collection<DataElement> elements) {
		if (elements.isEmpty()) {
			return false;
		}

		for (DataElement element : elements) {
			if (highlights.contains(element)) {
				highlights.remove(element);
			} else {
				highlights.add(element);
			}
		}

		fireUpdateListeners();
		return true;
	}


	/**
	 * @return All currently highlighted elements.
	 */
	public DataElement[] getHighlights() {
		return updateListeners.toArray(new DataElement[] {});
	}


	/**
	 * Sets which elements are currently highlighted.
	 *
	 * @param elements The elements to highlight.
	 * @return True if the set of highlighted elements was changed.
	 */
	public boolean setHighlights(Set<DataElement> elements) {
		Objects.requireNonNull(elements);

		if (highlights.equals(elements)) {
			return false;
		} else {
			this.highlights = elements;
			fireUpdateListeners();
			return true;
		}
	}


	/**
	 * Sets which element the cursor is currently touching.
	 *
	 * @param element The element below the cursor.
	 * @return True if the set of touched elements was changed.
	 */
	public boolean setTouched(DataElement element) {
		Objects.requireNonNull(element);

		if (touched.contains(element)) {
			return false;
		} else {
			touched.clear();
			touched.add(element);
			fireUpdateListeners();
			return true;
		}
	}


	/**
	 * Sets which elements the cursor is currently touching.
	 *
	 * @param elements The set of elements below the cursor.
	 * @return True if the set of touched elements was changed.
	 */
	public boolean setTouched(Set<DataElement> elements) {
		Objects.requireNonNull(elements);

		if (touched.equals(elements)) {
			return false;
		} else {
			this.touched = elements;
			fireUpdateListeners();
			return true;
		}
	}


	/**
	 * Removes all touches.
	 */
	public boolean clearTouched() {
		if (touched.isEmpty()) {
			return false;
		} else {
			touched.clear();
			fireUpdateListeners();
			return true;
		}
	}


	/**
	 * @param element The element to test.
	 * @return True if element is under the cursor in some view.
	 */
	public boolean isTouched(DataElement element) {
		return touched.contains(element);
	}


	@Override
	public boolean addUpdateListener(UpdateListener listener) {
		Objects.requireNonNull(listener);
		return updateListeners.add(listener);
	}


	@Override
	public boolean removeUpdateListener(UpdateListener listener) {
		return updateListeners.remove(listener);
	}


	@Override
	public UpdateListener[] getUpdateListeners() {
		return updateListeners.toArray(new UpdateListener[] {});
	}


	/**
	 * Fires all update listeners.
	 */
	public void fireUpdateListeners() {
		if (destroyed) {
			return;
		}

		for (UpdateListener listener : updateListeners) {
			try {
				listener.updatePerformed();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}


	@Override
	public boolean addDestroyListener(DestroyListener listener) {
		Objects.requireNonNull(listener);
		return destroyListeners.add(listener);
	}


	@Override
	public boolean removeDestroyListener(DestroyListener listener) {
		return destroyListeners.remove(listener);
	}


	@Override
	public DestroyListener[] getDestroyListeners() {
		return destroyListeners.toArray(new DestroyListener[] {});
	}


	@Override
	public void destroyPerformed() {
		if (destroyed) {
			return;
		}

		destroyed = true;

		DestroyListener[] listeners = destroyListeners.toArray(new DestroyListener[] {});

		for (DestroyListener listener : listeners) {
			try {
				listener.destroyPerformed();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}


	/**
	 * Indicates whether or not the object will continue emitting events.
	 *
	 * @return True if the destroyPerformed method has been called.
	 */
	public boolean isDestroyed() {
		return destroyed;
	}


	/**
	 * Used for detecting changes in selections.
	 */
	private class SelectionUpdateListener implements UpdateListener {
		@Override
		public void updatePerformed() {
			fireUpdateListeners();
		}
	}
}
