/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.data;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import vispaca.misc.UncheckedException;


/**
 * CSV parser.
 */
public class DataParser {

	public static final Pattern ATTRIBUTE_PATTERN = Pattern.compile("([0-9\\.]+|\"[^\"]+\")");


	/**
	 * @param stream The stream to read from.
	 * @return The parser data table.
	 */
	public static DataTable parse(InputStream stream) {
		List<String> rawRows;

		try {
			rawRows = getRawRows(stream);
		} catch (IOException e) {
			throw new UncheckedException(e);
		}

		if (rawRows.size() < 2) {
			throw new IllegalArgumentException("Input contains no elements.");
		}

		/* Parse and remove the header. */
		String[] attributeNames = getAttributeNames(rawRows.remove(0));

		/* Retrieve the types from a sample element. */
		Class<?>[] attributeTypes = getAttributeTypes(rawRows.get(0));

		/* Parse all elements and verify consistent types. */
		DataElement[] elements = getElements(rawRows, attributeTypes);

		return new DataTable(attributeTypes, attributeNames, elements);
	}


	/**
	 * @param stream The stream to read from.
	 * @return A list of raw table rows.
	 */
	private static List<String> getRawRows(InputStream stream) throws IOException {
		BufferedReader reader = new BufferedReader(new InputStreamReader(stream));

		List<String> rawRows = new ArrayList<>();
		String rawLine;

		while ((rawLine = reader.readLine()) != null) {
			rawLine = rawLine.trim();

			if (0 < rawLine.length()) {
				rawRows.add(rawLine);
			}
		}

		reader.close();

		return rawRows;
	}


	/**
	 * @return The names of all attributes.
	 */
	private static String[] getAttributeNames(String rawHeader) {
		List<CellTuple> cells = parseRow(rawHeader);
		int cellCount = cells.size();

		String[] names = new String[cellCount];

		for (int i = 0; i < cellCount; i++) {
			CellTuple cell = cells.get(i);

			if (cell.type != String.class) {
				throw new IllegalArgumentException("First row must contain only strings (attribute names)!");
			}

			names[i] = cell.data;
		}

		return names;
	}


	/**
	 * @return The types of all attributes.
	 */
	private static Class<?>[] getAttributeTypes(String rawElement) {
		List<CellTuple> cells = parseRow(rawElement);
		int cellCount = cells.size();
		Class<?>[] types = new Class<?>[cellCount];

		for (int i = 0; i < cellCount; i++) {
			types[i] = cells.get(i).type;
		}

		return types;
	}


	/**
	 * @return The elements of the table.
	 */
	private static DataElement[] getElements(List<String> rawRows, Class<?>[] attributeTypes) {
		int attributeCount = attributeTypes.length;
		int rowCount = rawRows.size();
		DataElement[] elements = new DataElement[rowCount];

		for (int i = 0; i < rowCount; i++) {
			List<CellTuple> cells = parseRow(rawRows.get(i));

			if (cells.size() != attributeCount) {
				throw new IllegalArgumentException("Inconsistent attribute count for element " + i + ".");
			}

			String[] attributes = new String[attributeCount];

			for (int j = 0; j < attributeCount; j++) {
				CellTuple cell = cells.get(j);

				if (cell.type != attributeTypes[j]) {
					throw new IllegalArgumentException("Inconsistent attribute type for element " + i + ", attribute " + j + ".");
				}

				attributes[j] = cell.data;
			}

			elements[i] = new DataElement(attributes);
		}

		return elements;
	}


	/**
	 * @param rawRow The raw row to parse.
	 * @return CellTuple array for the row.
	 */
	private static List<CellTuple> parseRow(String rawRow) {
		Matcher matcher = ATTRIBUTE_PATTERN.matcher(rawRow);
		List<CellTuple> cells = new ArrayList<>();

		while (matcher.find()) {
			CellTuple cell = new CellTuple();
			int start = matcher.start();
			int end = matcher.end();

			if (rawRow.charAt(start) == '"') {
				cell.data = rawRow.substring(start + 1, end - 1);
				cell.type = String.class;
			} else {
				cell.data = rawRow.substring(start, end);

				if (cell.data.contains(".")) {
					cell.type = Double.class;
					Double.parseDouble(cell.data);
				} else {
					cell.type = Integer.class;
					Integer.parseInt(cell.data);
				}
			}

			cells.add(cell);
		}

		return cells;
	}


	/**
	 * Internal data structure.
	 */
	private static class CellTuple {
		Class<?> type;
		String data;
	}
}
