/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.misc;

import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.geom.Rectangle2D;
import java.text.DecimalFormat;
import java.util.Objects;


/**
 * Draws rules in Swing for coordinate systems in 2D.
 */
public class ScatterRulerUtil {

	private Graphics g;
	private FontMetrics fontMetrics;

	private CoordinateMapper xMapper;
	private CoordinateMapper yMapper;

	private DecimalFormat decimalFormat = new DecimalFormat("0.0");


	public ScatterRulerUtil() {

	}


	public ScatterRulerUtil(Graphics g, CoordinateMapper xMapper, CoordinateMapper yMapper) {
		init(g, xMapper, yMapper);
	}


	/**
	 * For late initialization or reuse of objects.
	 */
	public void init(Graphics g, CoordinateMapper xMapper, CoordinateMapper yMapper) {
		Objects.requireNonNull(g);
		Objects.requireNonNull(xMapper);
		Objects.requireNonNull(yMapper);

		this.g = g;
		this.xMapper = xMapper;
		this.yMapper = yMapper;

		fontMetrics = g.getFontMetrics();

		/* Define the rulers sizes. */
		int rulerWidthMin = (int) fontMetrics.getStringBounds("-" + decimalFormat.format(yMapper.getValueMin()), g).getWidth() + 10;
		int rulerWidthMax = (int) fontMetrics.getStringBounds("-" + decimalFormat.format(yMapper.getValueMax()), g).getWidth() + 10;
		int rulerHeight = fontMetrics.getHeight() + 10;

		if (rulerWidthMin <= rulerWidthMax) {
			xMapper.setPixelMin(rulerWidthMax + 1);
		} else {
			xMapper.setPixelMin(rulerWidthMin + 1);
		}

		yMapper.setPixelMax(yMapper.getPixelMax() - rulerHeight - 1);
	}


	/**
	 * Draw the rulers.
	 */
	public void drawRulers(int yStepCount, int xStepCount) {
		g.setColor(ColorDefiner.NEUTRAL);

		yStepCount--;
		xStepCount--;

		double yStep = yMapper.getValueDelta() / yStepCount;
		double yValue = Math.min(yMapper.getValueMin(), yMapper.getValueMax());

		for (int i = 0; i < yStepCount; i++) {
			drawRulerTextY(yValue);
			yValue += yStep;
		}

		drawRulerTextY(Math.max(yMapper.getValueMin(), yMapper.getValueMax()));

		double xStep = xMapper.getValueDelta() / xStepCount;
		double xValue = Math.min(xMapper.getValueMin(), xMapper.getValueMax());

		for (int i = 0; i < xStepCount; i++) {
			drawRulerTextX(xValue);
			xValue += xStep;
		}

		drawRulerTextX(Math.max(xMapper.getValueMin(), xMapper.getValueMax()));
	}


	/**
	 * Draw the ruler for a value on the Y axis.
	 */
	public void drawRulerTextY(double value) {
		FontMetrics fontMetrics = g.getFontMetrics();
		String text = decimalFormat.format(value);
		Rectangle2D aText = fontMetrics.getStringBounds(text, g);
		int yCenter = (int) Math.round(yMapper.toPixel(value));
		int xStart = (int) Math.round(xMapper.getPixelMin() - 1);
		int xEnd = (int) (int) Math.round(xMapper.getPixelMax() - 1);

		int yText = yCenter - 1 + (int) (aText.getHeight() / 2);
		int xText = xStart - 6 - (int) aText.getWidth();

		if (yCenter <= aText.getHeight()) {
			yText += (aText.getHeight() / 2) + 1;
		}

		g.drawLine(xStart - 3, yCenter, xEnd, yCenter);
		g.drawString(text, xText, yText);
	}


	/**
	 * Draw the ruler for a value on the X axis.
	 */
	public void drawRulerTextX(double value) {
		FontMetrics fontMetrics = g.getFontMetrics();
		String text = decimalFormat.format(value);
		Rectangle2D aText = fontMetrics.getStringBounds(text, g);
		int xCenter = (int) Math.round(xMapper.toPixel(value));
		int yBottom = (int) yMapper.getPixelMax() + 1;
		int yTop = (int) Math.round(yMapper.getPixelMin());

		int xText = xCenter - (int) (aText.getWidth() / 2);
		int yText = yBottom + 6 + (int) aText.getHeight();

		if (xMapper.getPixelMax() <= aText.getWidth() + xCenter) {
			xCenter -= 1;
			xText -= (aText.getWidth() / 2) + 1;
		}

		g.drawLine(xCenter, yTop, xCenter, yBottom + 3);
		g.drawString(text, xText, yText);
	}
}
